@@ -38,3 +38,10 @@ test.nestedlist <- function()
 	checkIdentical( x, correct )
 	checkIdentical( x[[2]], correct[[2]] )
 }
+
+test.bad.list <- function()
+{
+	bad_json <- "{\"a\": 123,}"
+	x <- try( fromJSON( bad_json ), silent = TRUE )
+	checkTrue( any( class( x ) == "try-error" ) )
+}
